require 'redis'

class RedisBid

  class ItemNotRegisteredError < StandardError; end
  class InvalidBidScoreError < StandardError; end

  DEFAULT_USER_ID = 0
  DEFAULT_MIN_BID_SCORE   = 0

  BID_ITEMS_KEY_PREFIX = "bid_items"
  BID_ITEM_KEY_PREFIX  = "bid"

  def self.redis=(conn) @@redis = conn end
  def self.redis
    @@redis ||= Redis.new(:host => 'localhost', :port => 6379)
  end

  def self.register_bid_item(bid_id, options = {})
    options = { min_bid_score: DEFAULT_MIN_BID_SCORE }.merge(options)
    return false if options[:min_bid_score] < DEFAULT_MIN_BID_SCORE
    redis.sadd(BID_ITEMS_KEY_PREFIX, bid_id)
    set_min_bid(bid_id, options[:min_bid_score])
  end

  def self.unregister_bid_item(bid_id)
    raise ItemNotRegisteredError.new unless redis.srem(BID_ITEMS_KEY_PREFIX, bid_id)
    redis.hdel("min_bid", bid_id)
    return true
  end

  def self.is_registered?(bid_id)
    redis.sismember(BID_ITEMS_KEY_PREFIX, bid_id)
  end

  def self.set_min_bid(bid_id, score)
    raise ItemNotRegisteredError.new unless is_registered?(bid_id)
    raise InvalidBidScoreError.new("score should be >= #{DEFAULT_MIN_BID_SCORE}") if score < DEFAULT_MIN_BID_SCORE
    redis.hset("min_bid", bid_id, score)
  end

  def self.get_min_bid(bid_id)
    redis.hget("min_bid", bid_id).to_i
  end

  def self.bid_items
    redis.smembers(BID_ITEMS_KEY_PREFIX)
  end

  def self.add_user_bid(user_id, bid_id, score)
    raise ItemNotRegisteredError.new unless is_registered?(bid_id)
    min_bid_score = get_min_bid(bid_id)
    raise InvalidBidScoreError.new("score should be >= #{min_bid_score}") if score < min_bid_score
    redis.zadd("BID_ITEM_KEY_PREFIX:#{bid_id}", score, user_id)
  end

  def self.remove_user_bid(user_id, bid_id)
    redis.zrem("BID_ITEM_KEY_PREFIX:#{bid_id}", user_id)
  end

  def self.num_user_bids(bid_id)
    redis.zcard("BID_ITEM_KEY_PREFIX:#{bid_id}")
  end

  def self.list_user_bids(bid_id, options = {})
    options = { min_bid_score: DEFAULT_MIN_BID_SCORE, max_score: "+inf", with_scores: false, order: "desc"}.merge(options)
    return list_user_bids_in_descending_order(bid_id, options) if options[:order] == "desc"
    redis.zrangebyscore("BID_ITEM_KEY_PREFIX:#{bid_id}", options[:min_bid_score], options[:max_score], with_scores: options[:with_scores])
  end

  private

  def self.list_user_bids_in_descending_order(bid_id, options = {})
    redis.zrevrangebyscore("BID_ITEM_KEY_PREFIX:#{bid_id}", options[:max_score], options[:min_bid_score], with_scores: options[:with_scores])
  end
end
